library ac_dart.descriptor_model.hipchat;

import 'package:stuff/stuff.dart';
import 'package:atlassian_connect_host/src/descriptor/descriptor_model.dart';
import 'package:option/option.dart';
import 'package:quiver/iterables.dart';
import 'package:atlassian_connect_host/src/descriptor/hipchat/scopes.dart';

class _ProductlessScope extends Scope implements HipchatScope {
  _ProductlessScope(String key) : super(key);
}

class HipchatConnectAddon extends BaseConnectAddon<Capabilities> {
  final Capabilities capabilities;
  final HipchatLinks links;

  @override
  Capabilities get modules => capabilities;

  HipchatConnectAddon(
      {String key,
      String name,
      String description,
      this.capabilities,
      this.links})
      : super(key: key, name: name, description: description);

  factory HipchatConnectAddon.fromJson(Map json) {
    final p = parseJson(json);
    return new HipchatConnectAddon(
        key: p.single('key'),
        name: p.single('name'),
        description: p.single('description'),
        capabilities:
            p.single('capabilities', (v) => new Capabilities.fromJson(v)),
        links: p.single('links', (v) => new HipchatLinks.fromJson(v)));
  }

  Map toJson() => (buildJson
    ..add('key', key)
    ..add('name', name)
    ..add('description', description)
    ..add('capabilities', capabilities)
    ..add('links', links)).json;
}

class HipchatModuleTypes {
  static const String installable = 'installable';
  static const String configurable = 'configurable';
  static const String hipchatApiConsumer = 'hipchatApiConsumer';
  static const String oauth2Provider = 'oauth2Provider';
  static const String webhook = 'webhook';
}

class HipchatLinks implements Jsonable {
  final Uri homePage;
  final Uri self;

  HipchatLinks(this.self, {this.homePage});

  HipchatLinks._fromJson(JsonParser p)
      : this(p.single('self', (s) => Uri.parse(s)),
            homePage: p.single('homepage', (s) => Uri.parse(s)));

  HipchatLinks.fromJson(Map json) : this._fromJson(parseJson(json));

  Map toJson() =>
      (buildJson..add('self', self)..add('homepage', homePage)).json;
}

class Capabilities extends BaseModuleList {
  Option<Installable> get installable => single(HipchatModuleTypes.installable);
  Option<Configurable> get configurable =>
      single(HipchatModuleTypes.configurable);
  Option<HipchatApiConsumer> get hipchatApiConsumer =>
      single(HipchatModuleTypes.hipchatApiConsumer);
  Option<Oauth2Provider> get oauth2Provider =>
      single(HipchatModuleTypes.oauth2Provider);
  Iterable<HipchatWebhook> get webhooks => filtered(HipchatModuleTypes.webhook);

  Capabilities.fromModules(Iterable<ClassifiedModule> modules) : super(modules);

  static Iterable<ClassifiedModule> createCapabilities(
      Installable installable,
      Configurable configurable,
      HipchatApiConsumer hipchatApiConsumer,
      Oauth2Provider oauth2Provider,
      Iterable<Webhook> webhooks) {
    return concat([
      BaseModuleList.mapSingle(installable, HipchatModuleTypes.installable),
      BaseModuleList.mapSingle(configurable, HipchatModuleTypes.configurable),
      BaseModuleList.mapSingle(
          hipchatApiConsumer, HipchatModuleTypes.hipchatApiConsumer),
      BaseModuleList.mapSingle(
          oauth2Provider, HipchatModuleTypes.oauth2Provider),
      BaseModuleList.mapModules(webhooks, HipchatModuleTypes.webhook)
    ]);
  }

  Capabilities(
      {Installable installable,
      Configurable configurable,
      HipchatApiConsumer hipchatApiConsumer,
      Oauth2Provider oauth2Provider,
      Iterable<Webhook> webhooks})
      : this.fromModules(createCapabilities(installable, configurable,
            hipchatApiConsumer, oauth2Provider, webhooks));

  factory Capabilities.fromJson(Map json) {
    final p = parseJson(json);
    return new Capabilities(
        installable:
            p.single('installable', (v) => new Installable.fromJson(v)),
        configurable:
            p.single('configurable', (v) => new Configurable.fromJson(v)),
        hipchatApiConsumer: p.single(
            'hipchatApiConsumer', (v) => new HipchatApiConsumer.fromJson(v)),
        oauth2Provider:
            p.single('oauth2Provider', (v) => new Oauth2Provider.fromJson(v)),
        webhooks: p.list('webhooks', (v) => new HipchatWebhook.fromJson(v)));
  }

  Map toJson() => (buildJson
    ..add('installable', installable)
    ..add('configurable', configurable)
    ..add('hipchatApiConsumer', hipchatApiConsumer)
    ..add('oauth2Provider', oauth2Provider)
    ..add('webhooks', webhooks)).json;
}

class Installable implements Module {
  final bool allowGlobal;
  final bool allowRoom;
  final Uri callbackUri;

  Installable({this.allowGlobal, this.allowRoom, this.callbackUri});

  factory Installable.fromJson(Map json) {
    final p = parseJson(json);
    return new Installable(
        allowGlobal: p.single('allowGlobal'),
        allowRoom: p.single('allowRoom'),
        callbackUri: p.single('callbackUri', (v) => Uri.parse(v)));
  }

  Map toJson() => (buildJson
    ..add('allowGlobal', allowGlobal)
    ..add('allowRoom', allowRoom)
    ..add('callbackUri', callbackUri)).json;

  Installable copy({Uri callbackUri}) => new Installable(
      allowGlobal: allowGlobal,
      allowRoom: allowRoom,
      callbackUri: callbackUri != null ? callbackUri : this.callbackUri);
}

class Configurable implements Module {
  final bool allowAccessToRoomAdmins;
  final Uri url;

  Configurable(this.allowAccessToRoomAdmins, this.url);

  factory Configurable.fromJson(Map json) {
    final p = parseJson(json);
    return new Configurable(p.single('allowAccessToRoomAdmins'),
        p.single('url', (v) => Uri.parse(v)));
  }

  Map toJson() => (buildJson
    ..add('allowAccessToRoomAdmins', allowAccessToRoomAdmins)
    ..add('url', url)).json;

  Configurable copy({Uri url}) =>
      new Configurable(allowAccessToRoomAdmins, url != null ? url : this.url);
}

class HipchatApiConsumer implements Module {
  final Iterable<HipchatScope> scopes;

  HipchatApiConsumer(this.scopes);

  factory HipchatApiConsumer.fromJson(Map json) {
    final p = parseJson(json);
    return new HipchatApiConsumer(
        p.list('scopes', (v) => new _ProductlessScope(v)));
  }

  Map toJson() => (buildJson..add('scopes', scopes)).json;
}

class Oauth2Provider implements Module {
  final Uri tokenUrl;
  final Uri authorizationUrl;

  Oauth2Provider({this.authorizationUrl, this.tokenUrl});

  factory Oauth2Provider.fromJson(Map json) {
    final p = parseJson(json);
    return new Oauth2Provider(
        tokenUrl: p.single('tokenUrl', (s) => Uri.parse(s)),
        authorizationUrl: p.single('authorizationUrl', (s) => Uri.parse(s)));
  }

  Map toJson() => (buildJson
    ..add('tokenUrl', tokenUrl)
    ..add('authorizationUrl', authorizationUrl)).json;
}

class HipchatWebhook implements Module {
  final String event;
  final RegExp pattern;
  final Uri url;
  final String name;

  HipchatWebhook({this.event, this.pattern, this.url, this.name});

  factory HipchatWebhook.fromJson(Map json) {
    final p = parseJson(json);
    return new HipchatWebhook(
        event: p.single('event'),
        pattern: p.single('pattern', (s) => new RegExp(s)),
        url: p.single('url', (u) => Uri.parse(u)),
        name: p.single('name'));
  }

  Map toJson() => (buildJson
    ..add('event', event)
    ..add('pattern', pattern)
    ..add('url', url)
    ..add('name', name)).json;
}
