part of ac_dart.host_service.service;

class _AcHostServiceAuthentication {
  final AcHostRegistrationRepository _hostRegistrationRepository;
  final AddonHost _addonHost;
  final Set<JsonWebAlgorithm> _supportedAlgorithms;

  _AcHostServiceAuthentication(
      AcHostRegistrationRepository hostRegistrationRepository,
      AddonHost addonHost,
      {Set<JsonWebAlgorithm> supportedAlgorithms})
      : this._hostRegistrationRepository =
            checkNotNull(hostRegistrationRepository),
        this._addonHost = checkNotNull(addonHost),
        this._supportedAlgorithms = supportedAlgorithms != null
            ? supportedAlgorithms
            : new Set.from([JsonWebAlgorithm.HS256]);

  Future<Either<AuthenticationFailure, AcPrincipal>> authenticateHostRequest(
      String jwtToken, Uri uri, String method) {
    try {
      final JsonWebToken jwt = decodeProductHostToken(jwtToken);

      final claimSet = jwt.claimSet;

      return _authenticate(
          jwt,
          claimSet.issuer,
          (AcHostRegistration hostRego) =>
              _createValidationContext(uri, method, hostRego));
    } on ArgumentError catch (e) {
      return new Future.value(
          new Left(new AuthenticationFailure.singleError(400, e.message)));
    }
  }

  JwtValidationContext _createValidationContext(
      Uri uri, String method, AcHostRegistration hostRegistration) {
    final JwaSignatureContext signatureContext =
        new JwaSymmetricKeySignatureContext(hostRegistration.sharedSecret);

    final claimSetValidationContext = new ProductHostClaimSetValidationContext(
        uri, method,
        contextPath: _addonHost.baseUrl.path);

    return new JwtValidationContext(
        signatureContext, claimSetValidationContext);
  }

  @deprecated
  String createAddonSessionToken(String hostKey, String user) {
    if (_addonHost.addonSecret is None) {
      throw new ArgumentError('must configure a secret for the addon to '
          'create session tokens');
    }
    return session_token.createAddonSessionToken(
        _addonHost.addonSecret.get(), _addonHost.key, hostKey, user);
  }

  @deprecated
  Future<
      Either<AuthenticationFailure, AcPrincipal>> authenticateAddonSessionToken(
      String jwtToken) {
    try {
      final JsonWebToken jwt = decodeAddonSessionToken(jwtToken);
      final AddonSessionClaimSet claimSet = jwt.claimSet;

      final issuer = claimSet.issuer;

      if (issuer != _addonHost.key) {
        return new Future.value(new Left(new AuthenticationFailure.singleError(
            HttpStatus.UNAUTHORIZED,
            "The token was not issued by this addon")));
      }

      return _authenticate(
          jwt,
          claimSet.productHostKey,
          (_) => new JwtValidationContext.withSharedSecret(
              _addonHost.addonSecret.get()));
    } on ArgumentError catch (e) {
      return new Future.value(new Left(new AuthenticationFailure.singleError(
          400, 'invalid jwtToken: ${e.message}')));
    }
  }

  Future<Either<AuthenticationFailure, AcPrincipal>> _authenticate(
      JsonWebToken<OpenIdJwtClaimSet> jwt,
      String issuer,
      JwtValidationContext createValidationContext(
          AcHostRegistration hostRego)) async {
    _log.finest('authenticate claimSet: ${jwt.claimSet}');

    final header = (jwt as JsonWebSignature).header;
    if (!_supportedAlgorithms.contains(header.algorithm)) {
      return new Future.value(new Left(new AuthenticationFailure.singleError(
          HttpStatus.UNAUTHORIZED, "unsupported algorithm")));
    }

    final hostRegistrationOpt =
        await _hostRegistrationRepository.findRegistrationByHostKey(issuer);

    return hostRegistrationOpt.map((hostRegistration) {
      _log.finest('hostRego is $hostRegistration');

      Set<ConstraintViolation> violations =
          jwt.validate(createValidationContext(hostRegistration));

      return _createResult(violations, jwt.claimSet, hostRegistration);
    }).getOrElse(() => new Left(new AuthenticationFailure.singleError(
        HttpStatus.FORBIDDEN, "unknown host")));
  }

  Either<AuthenticationFailure, AcPrincipal> _createResult(
      Set<ConstraintViolation> violations,
      OpenIdJwtClaimSet claimSet,
      AcHostRegistration hostRegistration) {
    if (violations.isEmpty) {
      // bitbucket doesn't set the subject. So we establish a login with a
      // special subject that we can give limited access
      final subject =
          claimSet.subject != null ? claimSet.subject : acDartUnknownSubject;

      return new Right(
          new AcPrincipal(subject, hostRegistration.hostDetails.key));
    } else {
      return new Left(new AuthenticationFailure(HttpStatus.UNAUTHORIZED,
          new ConstraintViolations('Authentication failed', violations)));
    }
  }
}

class AuthenticationFailure {
  final int statusCode;
  final ConstraintViolations constraintViolations;

  String get message =>
      'Authentication Failed: ${constraintViolations.summaryMessage}';

  AuthenticationFailure(this.statusCode, this.constraintViolations);

  AuthenticationFailure.singleError(int statusCode, String message)
      : this(
            statusCode,
            new ConstraintViolations('Authentication failed',
                new Set()..add(new ConstraintViolation(message))));
}

// TODO: review where this came from
class HcPrincipal extends Principal {
  final int id;

  HcPrincipal(this.id, String name) : super(name);
}

const String acDartUnknownSubject = 'acdart:unknown';

class AcPrincipal extends Principal {
  final String hostKey;

  @deprecated // use name
  String get user => name;

  AcPrincipal(String name, this.hostKey) : super(name);
}

class AddonSessionAuthenticationHelper {
  final AcHostRegistrationRepository _hostRegistrationRepository;
  final AddonHost _addonHost;

  AddonSessionAuthenticationHelper(
      this._hostRegistrationRepository, this._addonHost) {
    checkNotNull(_addonHost);
    if (_addonHost.addonSecret is None) {
      throw new ArgumentError(
          'Must provide addonSecret to use addon session tokens');
    }
  }

  Future<AddonSessionClaimSet> _addonSessionClaimFactory(
      String issuer,
      AcPrincipal principal,
      String sessionIdentifier,
      Duration idleTimeout,
      Duration totalSessionTimeout) async {
    return new AddonSessionClaimSet(issuer, principal.name, principal.hostKey,
        sessionIdentifier: sessionIdentifier,
        idleTimeout: idleTimeout,
        totalSessionTimeout: totalSessionTimeout);
  }

  Future<Option<AcPrincipal>> _lookupUser(AddonSessionClaimSet claimSet) async {
    final issuer = claimSet.issuer;

    if (issuer != _addonHost.key) {
      _throwUnauthenticated("The token was not issued by this addon");
    }

    final hostRegistrationOpt = await _hostRegistrationRepository
        .findRegistrationByHostKey(claimSet.productHostKey);

    if (hostRegistrationOpt is None) {
      _throwForbidden("unknown host ${claimSet.productHostKey}");
    }

    return new Some(new AcPrincipal(
        claimSet.subject, hostRegistrationOpt.get().hostDetails.key));
  }

  JwtSessionHandler<AcPrincipal, AddonSessionClaimSet> get jwtSessionHandler =>
      new JwtSessionHandler.custom(_addonHost.key, _addonHost.addonSecret.get(),
          _lookupUser, _addonSessionClaimFactory,
          jwtCodec: addonSessionCodec);
}

_throwUnauthenticated(String message) {
  throw new UnauthorizedException({'error': message}, message);
}

_throwForbidden(String message) {
  throw new ForbiddenException({'error': message}, message);
}
