library ac_dart.host_service.auth.test;

import 'package:test/test.dart';
import 'package:mockito/mockito.dart';
import 'dart:async';

import 'package:atlassian_connect_jwt/atlassian_connect_jwt.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart';
import 'package:atlassian_connect_host/host_model.dart';
import 'package:logging/logging.dart';
import 'package:either/either.dart';
import 'package:atlassian_connect_host/host_registration_persistence.dart';
import 'package:path/path.dart' as _path;

//class _MockHttpRequest extends Mock implements HttpRequest {
//  noSuchMethod(i) => super.noSuchMethod(i);
//}
//
//class _MockHttpHeaders extends Mock implements HttpHeaders {
//  noSuchMethod(i) => super.noSuchMethod(i);
//}
//
//class _MockHttpResponse extends Mock implements HttpResponse {
//  noSuchMethod(i) => super.noSuchMethod(i);
//}

class _MockAcHostRegistrationRepository extends Mock
    implements AcHostRegistrationRepository {
  noSuchMethod(i) => super.noSuchMethod(i);
}

// HostClientFactory is a typedef

abstract class HostClientFactoryClass {
  AcHostHttpClient call(
      AcHostRegistration hostRegistration, AddonHost addonHost);
}

class _MockHostClientFactory extends Mock implements HostClientFactoryClass {
  noSuchMethod(i) => super.noSuchMethod(i);
}

void _initLogging() {
  Logger.root.level = Level.WARNING;
  Logger.root.onRecord.listen((LogRecord rec) {
    print('${rec.message}');
  });
}

final p = _path.url;

void main() {
  _initLogging();
  const String addonKey = 'my-addon';
  final String addonContext = 'foo';
  final Uri addonBaseUri = Uri.parse('http://localhost:9999/$addonContext');
  final String hostKey = "jira:ae390d29-31b2-4c12-a719-9df64e3e92b7";
  final String sharedSecret = '3ab90b11-d7bd-4097-958f-01b7ac4e985f';
  const String publicKey = '93475ljsdrt93kwd';
  const String username = 'fred';

  final _hostRegistration = (new AcHostRegistration.build(
      sharedSecret: "$sharedSecret",
      hostDetails: (new AcHostDetails.build(
          key: "$hostKey",
          baseUrl: Uri.parse('http://localhost:8080/blah'),
          productType: ProductType.jira,
          publicKey: "$publicKey"))));

  _MockAcHostRegistrationRepository _registrationRepository;
  _MockHostClientFactory _hostClientFactory;

  final String requestPath = 'hello-world';
  final String uriStrNoJwt =
      "$requestPath?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&xdm_c=channel-servlet-hello-world&xdm_p=1";
  final Uri uriNoJwt = Uri.parse(p.join(p.fromUri(addonBaseUri), uriStrNoJwt));
  final String invalidJwtStr =
      r'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjEzOTQxNjYzNjcsInN1YiI6ImFkbWluIiwiaXNzIjoiamlyYTphZTM5MGQyOS0zMWIyLTRjMTItYTcxOS05ZGY2NGUzZTkyYjciLCJxc2giOiJlYjRlYzVmMDk4MGQwMWRhMzA2ZGI4OWZhNTdkZDE2MzU3NDY2NzQzNmRkNzIwZDVkOWM1Mjc5YzViNDVmN2E4IiwiaWF0IjoxMzk0MTY2MTg3fQ.bR8Z0MIguOg6xgHiav0quun8kTqXzNUKMKym-PFjZvc';
  final String uriStrInvalidJwt =
      "$requestPath?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&xdm_c=channel-servlet-hello-world&xdm_p=1&jwt=${invalidJwtStr}";
  final Uri uriInvalidJwt =
      Uri.parse(p.join(p.fromUri(addonBaseUri), uriStrInvalidJwt));

//  final hostRego = (new AcHostRegistration.build(
//      sharedSecret: "$sharedSecret",
//      hostDetails: (new AcHostDetails.build(
//          key: "$hostKey",
//          baseUrl: Uri.parse('http://localhost:8080'),
//          productType: ProductType.jira,
//          publicKey: "pubby"))));

  String jwtToken = createProductHostToken(
      requestPath, 'GET', uriNoJwt.queryParameters, sharedSecret, hostKey);

  final Uri uriWithJwt =
      Uri.parse(p.join(p.fromUri(addonBaseUri), '$uriStrNoJwt&jwt=$jwtToken'));

  setUp(() {
    _registrationRepository = new _MockAcHostRegistrationRepository();
    when(_registrationRepository.findRegistrationByHostKey(any))
        .thenReturn(new Future.value(new Some(_hostRegistration)));
    _hostClientFactory = new _MockHostClientFactory();
  });

  group('[host authentication]', () {
    AcHostService hostService;

    setUp(() {
      final AddonHost addonHost =
          new AddonHost(addonKey, addonBaseUri, 'sseecret');

      hostService = new AcHostService(
          _registrationRepository,
          addonHost,
          (hostRego, addonHost) =>
              _hostClientFactory.call(hostRego, addonHost));
    });

    Future<Either<AuthenticationFailure, AcPrincipal>> authenticate(Uri uri) {
      final jwtToken = uri.queryParameters['jwt'];
      return hostService.authenticateHostRequest(jwtToken, uri, 'GET');
    }

    _expect(Uri uri, int expectedStatusCode) async {
      final Either<AuthenticationFailure, AcPrincipal> result =
          await authenticate(uri);
      expect(result.isRight(), equals(expectedStatusCode == null));

      if (expectedStatusCode != null) {
        expect(result.left().get().statusCode, equals(expectedStatusCode));
      }
    }

    _expectOccurs(Uri uri,
        void check(Either<AuthenticationFailure, AcPrincipal> result)) async {
      final Either<AuthenticationFailure, AcPrincipal> result =
          await authenticate(uri);

      check(result);
    }

    group('[invalid jwt]', () {
      test('looks up host registration when jwt in request', () {
        _expectOccurs(
            uriInvalidJwt,
            (_) =>
                verify(_registrationRepository.findRegistrationByHostKey(any))
                    .called(1));
      });

      test('looks up host registration with correct issuer when jwt in request',
          () {
        _expectOccurs(
            uriInvalidJwt,
            (_) =>
                verify(_registrationRepository.findRegistrationByHostKey(any))
                    .called(1));
      });

      test('sets status code to 401 when invalid jwt in request', () {
        _expect(uriInvalidJwt, 401);
      });
    });

    group('[valid jwt]', () {
      test('Doesnt return error code when valid jwt in request', () {
        _expect(uriWithJwt, null);
      });

      test('looks up host registration when jwt in request', () {
        _expectOccurs(
            uriWithJwt,
            (_) =>
                verify(_registrationRepository.findRegistrationByHostKey(any))
                    .called(1));
      });

      test('looks up host registration with correct issuer when jwt in request',
          () {
        _expectOccurs(
            uriWithJwt,
            (_) => verify(
                    _registrationRepository.findRegistrationByHostKey(hostKey))
                .called(1));
      });
    });

    group('[no such issuer]', () {
      setUp(() {
        when(_registrationRepository.findRegistrationByHostKey(any))
            .thenReturn(new Future.value(const None()));
      });

      test('Authentication fails', () {
        _expectOccurs(uriWithJwt, (result) => result.isLeft());
      });

      test('status code is 403', () {
        _expect(uriWithJwt, 403);
      });
    });
  });

  group('[addon session token]', () {
    const String addonSecret = "sssshhhhh";
    const String unknownAddonSecret = 'something completely different';

    AddonHost addonHost(String secret) =>
        new AddonHost(addonKey, addonBaseUri, secret);

    AcHostService hostService(String secret) => new AcHostService(
        _registrationRepository,
        addonHost(secret),
        (hostRego, addonHost) => _hostClientFactory.call(hostRego, addonHost));

    String sessionToken(String secret) =>
        hostService(secret).createAddonSessionToken(hostKey, username);

    AddonSessionClaimSet claimSet(String secret) =>
        decodeAddonSessionToken(sessionToken(addonSecret)).claimSet;

    group('[token creation]', () {
      test('throws ArgumentError if no addon secret', () {
        expect(
            () => hostService(null).createAddonSessionToken(hostKey, username),
            throwsArgumentError);
      });

      test('issuer is addon', () {
        expect(claimSet(addonSecret).issuer, equals(addonKey));
      });

      test('subject is user', () {
        expect(claimSet(addonSecret).subject, equals(username));
      });

      test('hostKey is correct', () {
        expect(claimSet(addonSecret).productHostKey, equals(hostKey));
      });

      test('hostKey is correct', () {
        expect(claimSet(addonSecret).productHostKey, equals(hostKey));
      });
    });

    group('[token authentication]', () {
      Future<Either<AuthenticationFailure, AcPrincipal>> authenticate(
          String jwtToken) {
        return hostService(addonSecret).authenticateAddonSessionToken(jwtToken);
      }

      void _expect(String jwtToken, int expectedStatusCode) {
        final authResultFuture = authenticate(jwtToken);
        authResultFuture
            .then((Either<AuthenticationFailure, AcPrincipal> result) {
          expect(result.isRight(), equals(expectedStatusCode == null));
          if (expectedStatusCode != null) {
            expect(result.left().get().statusCode, equals(expectedStatusCode));
          }
        });
        expect(authResultFuture, completes);
      }

      void _expectOccurs(String jwtToken,
          void check(Either<AuthenticationFailure, AcPrincipal> result)) {
        final authResultFuture = authenticate(jwtToken);
        authResultFuture
            .then((Either<AuthenticationFailure, AcPrincipal> result) {
          check(result);
        });
        expect(authResultFuture, completes);
      }

      group('[invalid jwt]', () {
        test('sets status code to 400 when invalid jwt in request', () {
          _expect(invalidJwtStr, 400);
        });

        test('looks up host registration when jwt in request', () {
          _expectOccurs(
              sessionToken(unknownAddonSecret),
              (_) =>
                  verify(_registrationRepository.findRegistrationByHostKey(any))
                      .called(1));
        });

        test(
            'looks up host registration with correct issuer when jwt in request',
            () {
          _expectOccurs(
              sessionToken(unknownAddonSecret),
              (_) => verify(_registrationRepository
                  .findRegistrationByHostKey(hostKey)).called(1));
        });

        test('sets status code to 401 when invalid jwt in request', () {
          _expect(sessionToken(unknownAddonSecret), 401);
        });
      });

      group('[valid jwt]', () {
        test('returns Right when valid jwt in request', () {
          _expect(sessionToken(addonSecret), null);
        });

        test('looks up host registration when jwt in request', () {
          _expectOccurs(
              sessionToken(addonSecret),
              (_) =>
                  verify(_registrationRepository.findRegistrationByHostKey(any))
                      .called(1));
        });

        test(
            'looks up host registration with correct issuer when jwt in request',
            () {
          _expectOccurs(
              sessionToken(addonSecret),
              (_) => verify(_registrationRepository
                  .findRegistrationByHostKey(hostKey)).called(1));
        });

        test('session context has correct user', () {
          _expectOccurs(sessionToken(addonSecret), (result) {
            expect(result.right().get().user, equals(username));
          });
        });

        test('session context has correct product hostKey', () {
          _expectOccurs(sessionToken(addonSecret), (result) {
            expect(result.right().get().hostKey, equals(hostKey));
          });
        });
      });

      group('[no such issuer]', () {
        setUp(() {
          when(_registrationRepository.findRegistrationByHostKey(any))
              .thenReturn(new Future.value(const None()));
        });

        test('Authentication fails', () {
          _expectOccurs(sessionToken(addonSecret), (result) => result.isLeft());
        });

        test('status code is 403', () {
          _expect(sessionToken(addonSecret), 403);
        });
      });
    });
  });
}
